﻿// Ibis.cpp
// 2013 Phoenix

#include "Ibis.h"

HINSTANCE g_hDLL;

//=====================================================================================================================
//=====================================================================================================================
DLLCLBK void InitModule(HINSTANCE hModule)
{
  g_hDLL=hModule;
  
  // Perform global module initialisation here.
}

DLLCLBK void ExitModule(HINSTANCE hModule)
{
  // Perform module cleanup here.
}

//=====================================================================================================================
//=====================================================================================================================
DLLCLBK VESSEL * ovcInit(OBJHANDLE hVessel, int FlightModel)
{ 
  return new Ibis(hVessel, FlightModel);
}

DLLCLBK void ovcExit(VESSEL *pVessel)
{
  if(pVessel)
    delete (Ibis *)pVessel;
}

//=====================================================================================================================
//=====================================================================================================================

// Constructor.
Ibis::Ibis(OBJHANDLE hVessel, int FlightModel)
  : VESSEL3(hVessel, FlightModel), m_hPropellantTank(NULL)
{
  for(int t=0; t<NUM_THRUSTERS; t++)
  {
    m_hThruster[t]=NULL;
    m_ThrusterDeflection[t]=m_ReqThrusterDeflection[t]=0;
  }

  // Set the visual properties.
  SetSize(16);
  AddMesh(oapiLoadMeshGlobal("Ibis"));

  DefineAnimations();

  // Delta-v annotation. Changes every frame.
  m_hAnnotation[0]=oapiCreateAnnotation(true, 0.65, _V(0.3, 0.9, 0.3));
  oapiAnnotationSetPos(m_hAnnotation[0], 0.18, 0.1, 0.5, 0.14);

  // Name, bulk density and ISP annotation.
  m_hAnnotation[1]=oapiCreateAnnotation(true, 0.65, _V(0.3, 0.9, 0.3));
  oapiAnnotationSetPos(m_hAnnotation[1], 0.18, 0.13, 0.5, 0.20);
}
    
// Destructor.
Ibis::~Ibis()
{
  oapiDelAnnotation(m_hAnnotation[0]);
  oapiDelAnnotation(m_hAnnotation[1]); 
}

// Set vessel capabilities.
void Ibis::clbkSetClassCaps(FILEHANDLE hConfiguration)
{
  // Set the inertial properties of the rocket.
  SetEmptyMass(EMPTY_MASS);
  SetPMI(_V(4.46, 4.46, 0.18)); // Values from shipedit.

  // Set the touchdown points for a tail-sitter.
  SetTouchdownPoints(_V(0, 0.6, -3), _V(0.6, 0, -3), _V(-0.6, 0, -3));

  // Set the aerodynamic properties.
  SetCrossSections(_V(9.5, 9.5, 1.23)); // Values from shipedit.
  SetRotDrag(_V(0.1, 0.1, 0.005));
  SetCW(0.2, 1, 4, 4);

  // Initialise the acceleration at launch to 1.2Gs.
  double InitialAcceleration(INITIAL_ACCELERATION);

  // Read overriding acceleration configuration from the Ibis.cfg file.
  oapiReadItem_float(hConfiguration, "INITIAL_ACCELERATION", InitialAcceleration);

  // Initialise the rocket for thrust vector oscillation.
  m_MaxOscAmp=MAX_OSC_AMP;
  
  // Read overriding thrust vector oscillation.
  oapiReadItem_float(hConfiguration, "MAX_OSC_AMP", m_MaxOscAmp);

  // Initialise the rocket for hydrogen peroxide & kerosene propellant.
  char PropName[MAX_STRING]="Kerosene & hydrogen peroxide";
  double BulkDensity(1300), VacuumISP(319), SeaLevelISP(273);
  
  // Read overriding propellant configuration.
  oapiReadItem_string(hConfiguration, "PROPELLANT_NAME", PropName);
  oapiReadItem_float(hConfiguration, "BULK_DENSITY", BulkDensity);  // kilograms per cubic metre
  oapiReadItem_float(hConfiguration, "VACUUM_ISP", VacuumISP);      // seconds      
  oapiReadItem_float(hConfiguration, "SEA_LEVEL_ISP", SeaLevelISP);
  
  // Set the propellant name, bulk density and ISP annotation. This is only set once.
  char Annotation[MAX_STRING];
  sprintf_s(Annotation, MAX_STRING*sizeof(char), "Name = %s.\nDensity = %.0fkg/m3.\nISP = %.0fs (%.0fs).",
    PropName, BulkDensity, VacuumISP, SeaLevelISP);
  oapiAnnotationSetText(m_hAnnotation[1], Annotation);

  // Convert ISPs to metres per second.
  VacuumISP*=G;
  m_ExhaustVelocity=VacuumISP; // For calculating the delta-v capacity of the rocket.
  SeaLevelISP*=G;

  // Calculate propellant mass for the fixed tank volume and create the propellant resource.
  const double PropellantMass(PROP_TANK_VOLUME*BulkDensity);
  m_hPropellantTank=CreatePropellantResource(PropellantMass);

  // Calculate per thruster Newtons to give a sea-level (Earth pressure) initial acceleration at lift-off.
  const double
    ThrusterNewtons(((EMPTY_MASS+PropellantMass)*G*InitialAcceleration*(VacuumISP/SeaLevelISP))/NUM_THRUSTERS);

  // Add the thrusters and exhaust visuals.
  m_hThruster[THRUSTER_A]=
    CreateThruster(_V(0, 0.32, -2.516), _V(0, 0, 1), ThrusterNewtons, m_hPropellantTank, VacuumISP, SeaLevelISP);
  m_hThruster[THRUSTER_B]=
    CreateThruster(_V(0.32, 0, -2.516), _V(0, 0, 1), ThrusterNewtons, m_hPropellantTank, VacuumISP, SeaLevelISP);
  m_hThruster[THRUSTER_C]=
    CreateThruster(_V(0, -0.32, -2.516), _V(0, 0, 1), ThrusterNewtons, m_hPropellantTank, VacuumISP, SeaLevelISP);
  m_hThruster[THRUSTER_D]=
    CreateThruster(_V(-0.32, 0, -2.516), _V(0, 0, 1), ThrusterNewtons, m_hPropellantTank, VacuumISP, SeaLevelISP);

  EXHAUSTSPEC ExhaustSpec;
  memset(&ExhaustSpec, 0, sizeof(EXHAUSTSPEC));
  ExhaustSpec.lsize=8;
  ExhaustSpec.wsize=0.3;
  ExhaustSpec.lofs=0.6;
  ExhaustSpec.modulate=0.3;

  for(int t=THRUSTER_A; t<NUM_THRUSTERS; t++)
  {
    ExhaustSpec.th=m_hThruster[t];
    AddExhaust(&ExhaustSpec);
  }
  
  PARTICLESTREAMSPEC ParticleStream=
  {
    0,      // Flags.
    1,      // Particle size at creation.
    12,     // Particle emission rate in Hz.
    20.0,   // Initial velocity.
    0.1,    // Velocity spread during creation.
    2,      // Average particle lifetime.
    0.7,    // Growth rate.
    0.7,    // Slowdown rate in atmosphere.
    PARTICLESTREAMSPEC::DIFFUSE,
    PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.05,
    PARTICLESTREAMSPEC::ATM_PLIN, 1e-6, 1,
    NULL    // Texture.
  };

  // The exhaust streams are deliberately set to emerge some distance aft of the thrusters.
  AddExhaustStream(m_hThruster[THRUSTER_A], _V(0, 0.32, -16), &ParticleStream);
  AddExhaustStream(m_hThruster[THRUSTER_B], _V(0.32, 0, -16), &ParticleStream);
  AddExhaustStream(m_hThruster[THRUSTER_C], _V(0, -0.32, -16), &ParticleStream);
  AddExhaustStream(m_hThruster[THRUSTER_D], _V(-0.32, 0, -16), &ParticleStream);

  // Create MAIN thruster group. Note that this group controls roll, pitch and yaw, through vectoring thrust, as well
  // as forward propulsion. HUD autopilots do not work.
  CreateThrusterGroup(m_hThruster, NUM_THRUSTERS, THGROUP_MAIN);
}

// Handle keyboard.
int Ibis::clbkConsumeDirectKey(char *pKeyState)
{
  // Immediately return if playing back a recording.
  if(Playback())
    return 0;

  // Reset requested thrust vector deflections.
  for(int t=THRUSTER_A; t<NUM_THRUSTERS; t++)
      m_ReqThrusterDeflection[t]=0;

  // Set roll thrust vectoring.
  if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD4))    // Roll to port.
  {
    m_ReqThrusterDeflection[THRUSTER_A]=-MAX_DEFLECTION;
    m_ReqThrusterDeflection[THRUSTER_B]=MAX_DEFLECTION;
    m_ReqThrusterDeflection[THRUSTER_C]=MAX_DEFLECTION;
    m_ReqThrusterDeflection[THRUSTER_D]=-MAX_DEFLECTION;
  }
  else
    if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD6))  // Roll to starboard.
    {
      m_ReqThrusterDeflection[THRUSTER_A]=MAX_DEFLECTION;
      m_ReqThrusterDeflection[THRUSTER_B]=-MAX_DEFLECTION;
      m_ReqThrusterDeflection[THRUSTER_C]=-MAX_DEFLECTION;
      m_ReqThrusterDeflection[THRUSTER_D]=MAX_DEFLECTION;
    }

  // Set pitch vectoring.
  if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD8))    // Pitch down.
  {
    m_ReqThrusterDeflection[THRUSTER_B]=MAX_DEFLECTION;
    m_ReqThrusterDeflection[THRUSTER_D]=MAX_DEFLECTION;
  }
  else
    if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD2))  // Pitch up.
    {
      m_ReqThrusterDeflection[THRUSTER_B]=-MAX_DEFLECTION;
      m_ReqThrusterDeflection[THRUSTER_D]=-MAX_DEFLECTION;
    }

  // Set yaw vectoring.
  if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD1))    // Yaw to port.
  {
    m_ReqThrusterDeflection[THRUSTER_A]=MAX_DEFLECTION;
    m_ReqThrusterDeflection[THRUSTER_C]=MAX_DEFLECTION;
  }
  else
    if(KEYDOWN(pKeyState, OAPI_KEY_NUMPAD3))  // Yaw to starboard.
    {
      m_ReqThrusterDeflection[THRUSTER_A]=-MAX_DEFLECTION;
      m_ReqThrusterDeflection[THRUSTER_C]=-MAX_DEFLECTION;
    }

  return 0;
}

// Time step notification before state update. 
void Ibis::clbkPreStep(double SimT, double SimDT, double MJD)
{
  // Calculate the remaining delta-v capacity of the rocket using Tsiolkovsky's rocket equation, and update the
  // annotation.
  static char Annotation[MAX_STRING];
  
  sprintf_s(Annotation, sizeof(char)*MAX_STRING, "dV = %.0fm/s.",
    m_ExhaustVelocity*log((EMPTY_MASS+GetPropellantMass(m_hPropellantTank))/EMPTY_MASS));
  
  oapiAnnotationSetText(m_hAnnotation[0], Annotation);

  // Record the requested thruster deflections.
  if(Recording())
    if(m_RecTime<=0)
    {
      m_RecTime=1.0/RECORD_FREQ;  // Reset the recording timer.
    
      static char EventData[MAX_STRING];

      sprintf_s(EventData, sizeof(char)*MAX_STRING, "%+0.3f %+0.3f %+0.3f %+0.3f",
        m_ReqThrusterDeflection[THRUSTER_A], m_ReqThrusterDeflection[THRUSTER_B],
        m_ReqThrusterDeflection[THRUSTER_C], m_ReqThrusterDeflection[THRUSTER_D]);

      RecordEvent("REQTD", EventData);
    }
    else
      m_RecTime-=SimDT; // Continue countdown until the next sample.
  else
    m_RecTime=0;

  // Move thruster deflections to the requested angles.
  for(int t=THRUSTER_A; t<NUM_THRUSTERS; t++)
  {
    if(m_ThrusterDeflection[t]<m_ReqThrusterDeflection[t])
    {
      m_ThrusterDeflection[t]+=DEFLECTION_RATE*SimDT;
    
      if(m_ReqThrusterDeflection[t]==0 && m_ThrusterDeflection[t]>0)
        // This stops thruster oscillation either side of zero angle.
        m_ThrusterDeflection[t]=0;
      else
        if(m_ThrusterDeflection[t]>MAX_DEFLECTION)
          m_ThrusterDeflection[t]=MAX_DEFLECTION;
    }
    else
      if(m_ThrusterDeflection[t]>m_ReqThrusterDeflection[t])
      {
        m_ThrusterDeflection[t]-=DEFLECTION_RATE*SimDT;
      
        if(m_ReqThrusterDeflection[t]==0 && m_ThrusterDeflection[t]<0)
          // This stops thruster oscillation either side of zero angle.
          m_ThrusterDeflection[t]=0;
        else
          if(m_ThrusterDeflection[t]<-MAX_DEFLECTION)
            m_ThrusterDeflection[t]=-MAX_DEFLECTION;
      }
    
    // Compute a random oscillation to the thrust vector to create an animated effect to the exhaust flame. This will
    // also impart a little angular velocity to the rocket. 
    const double OscLng((oapiRand()-0.5)*m_MaxOscAmp), OscLat((oapiRand()-0.5)*m_MaxOscAmp);
    const VECTOR3 OscVec(_V(sin(OscLat)*cos(OscLng), sin(OscLat), cos(OscLng)*cos(OscLat)));

    // Set thruster direction.
    if(t & 1)
    {
      // Thruster B or D
      VECTOR3 ThVec(_V(OscVec.x, OscVec.y+sin(m_ThrusterDeflection[t]), OscVec.z+cos(m_ThrusterDeflection[t])));
      
      normalise(ThVec);
      SetThrusterDir(m_hThruster[t], ThVec);
    }
    else
    {
      // THRUSTER A or C
      VECTOR3 ThVec(_V(OscVec.x+sin(m_ThrusterDeflection[t]), OscVec.y, OscVec.z+cos(m_ThrusterDeflection[t])));

      normalise(ThVec);
      SetThrusterDir(m_hThruster[t], ThVec);
    }
    
    // Update nozzle mesh animation.
    SetAnimation(m_NozzleAnim[t], 0.5+(m_ThrusterDeflection[t]/MAX_DEFLECTION/2));
  }
}

// Recording playback.
bool Ibis::clbkPlaybackEvent(double SimT, double EventT, const char *pEventType, const char *pEventData)
{
  // Retrieve thruster deflection requests from the recording.
  if(strcmp(pEventType, "REQTD")!=0)
    return false;

  sscanf(pEventData, "%lf %lf %lf %lf", &m_ReqThrusterDeflection[THRUSTER_A], &m_ReqThrusterDeflection[THRUSTER_B],
    &m_ReqThrusterDeflection[THRUSTER_C], &m_ReqThrusterDeflection[THRUSTER_D]);

  return true;
}

void Ibis::DefineAnimations()
{
  // Thruster A
  static UINT GroupA[1]={ 3 }; // Participating groups

  static MGROUP_ROTATE ThrusterA(0,                           // Mesh index
                         GroupA, 1,                           // Group list and # groups 
                         _V(0, 0.32, -2.516),                 // Rotation reference point 
                         _V(0, 1, 0),                         // Rotation axis
                         static_cast<float>(MAX_DEFLECTION*2) // Angular rotation range
                         );

  m_NozzleAnim[THRUSTER_A]=CreateAnimation(0.5);
  AddAnimationComponent(m_NozzleAnim[THRUSTER_A], 0, 1, &ThrusterA);

  // Thruster B
  static UINT GroupB[1]={ 4 };
  static MGROUP_ROTATE
    ThrusterB(0, GroupB, 1, _V(0.32, 0, -2.516), _V(-1, 0, 0), static_cast<float>(MAX_DEFLECTION*2));

  m_NozzleAnim[THRUSTER_B]=CreateAnimation(0.5);
  AddAnimationComponent(m_NozzleAnim[THRUSTER_B], 0, 1, &ThrusterB);

  // Thruster C
  static UINT GroupC[1]={ 5 };
  static MGROUP_ROTATE 
    ThrusterC(0, GroupC, 1, _V(0, -0.32, -2.516), _V(0, 1, 0), static_cast<float>(MAX_DEFLECTION*2));

  m_NozzleAnim[THRUSTER_C]=CreateAnimation(0.5);
  AddAnimationComponent(m_NozzleAnim[THRUSTER_C], 0, 1, &ThrusterC);

  // Thruster D
  static UINT GroupD[1]={ 6 };
  static MGROUP_ROTATE
    ThrusterD(0, GroupD, 1, _V(-0.32, 0, -2.516), _V(-1, 0, 0), static_cast<float>(MAX_DEFLECTION*2));
  
  m_NozzleAnim[THRUSTER_D]=CreateAnimation(0.5);
  AddAnimationComponent(m_NozzleAnim[THRUSTER_D], 0, 1, &ThrusterD);
}