// Ibis.h
// 2013 Phoenix
//
// My first vessel addon for Orbiter Spaceflight Simulator. A compact, single-stage, four-engined, thrust-vectored,
// hydrogen-peroxide (H2O2) and kerosene rocket. This vessel can be reconfigured to use other propellants via the
// Ibis.cfg file. Four values are required to define a propellant: name, bulk density, vacuum ISP and sea-level ISP.
// The propellant tank volume and empty mass are fixed.

#define ORBITER_MODULE
#include "orbitersdk.h"

#ifndef IBIS_H
#define IBIS_H

// Vessel's physical parameters.
#define PROP_TANK_VOLUME (PI*0.608*0.608*4.75)  // Propellant tank volume in cubic metres.
#define EMPTY_MASS 800                          // Mass of the vessel without propellant in kilograms.
#define INITIAL_ACCELERATION 1.2                // Desired sea-level acceleration at lift-off in Gs.
#define MAX_DEFLECTION 0.12                     // Maximum positive and negative thruster deflection in radians.
#define DEFLECTION_RATE (MAX_DEFLECTION*2)      // Thruster deflection change rate in radians per second.
#define MAX_OSC_AMP 0.02                        // Maximum amplitude of random thrust vector oscillation in radians.

// Other constants.
#define RECORD_FREQ 30      // Maximum number of record samples in Hz.
#define MAX_STRING 96

class Ibis: public VESSEL3 
{
  public:
    // Constructor.
    Ibis(OBJHANDLE hVessel, int FlightModel=1);
    
    // Destructor.
    ~Ibis();

    // Set vessel capabilities.
    void clbkSetClassCaps(FILEHANDLE hConfiguration);

    // Handle keyboard.
    int clbkConsumeDirectKey(char *pKeyState);

    // Time step notification before state update. 
    void clbkPreStep(double SimT, double SimDT, double MJD);

    // Recording playback.
    bool clbkPlaybackEvent(double SimT, double EventT, const char *pEventType, const char *pEventData);   

    // Define engine nozzle deflection animations.
    void DefineAnimations();

  private:
    double m_RecTime; // Seconds until next record sample.

    PROPELLANT_HANDLE m_hPropellantTank;
    
    // Vectored thrusters. Thrusters A and C rotate along the y-axis, and thrusters B and D rotate along the x-axis.
    enum Thruster
    { 
      THRUSTER_A, THRUSTER_B, THRUSTER_C, THRUSTER_D, NUM_THRUSTERS
    };

    THRUSTER_HANDLE m_hThruster[NUM_THRUSTERS];

    // Current thruster deflections in radians, requested deflections, and the maximum thruster oscillation amplitude.
    double m_ThrusterDeflection[NUM_THRUSTERS], m_ReqThrusterDeflection[NUM_THRUSTERS], m_MaxOscAmp;

    UINT m_NozzleAnim[NUM_THRUSTERS];

    // Annotations.
    NOTEHANDLE m_hAnnotation[2];
    double m_ExhaustVelocity;
};

#endif // IBIS_H